# =============================================================================
# hardware_audit.py - plugin for auditing hardware for migrating classic XR to eXR/fleXR
#
# Copyright (c)  2013, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import re
import json

from csmpe.plugins import CSMPlugin
from migration_lib import SUPPORTED_HW_JSON, log_and_post_status, parse_admin_show_platform
from csmpe.core_plugins.csm_get_inventory.ios_xr.plugin import get_package, get_inventory


class Plugin(CSMPlugin):
    """
    A plugin for auditing hardware for migration from
    ASR9K IOS-XR (a.k.a. XR) to ASR9K IOS-XR 64 bit (a.k.a. eXR)

    Console access is needed.
    """
    name = "Migration Audit Plugin"
    platforms = {'ASR9K'}
    phases = {'Migration-Audit'}

    rp_pattern = re.compile('\d+/RS??P\d+/CPU\d+')
    fc_pattern = re.compile('\d+/FC\d+/SP')
    fan_pattern = re.compile('\d+/FT\d+/SP')
    pem_pattern = re.compile('\d+/P[SM]\d+/M?\d+/SP')
    lc_pattern = re.compile('\d+/\d+/CPU\d+')

    def _check_if_hw_supported_and_in_valid_state(self, inventory, supported_hw, override):
        """
        Check if RSP/RP/FAN/PEM/FC/LC/MPA currently on device are supported and are in valid state for migration.

        Minimal requirements (with override=True):
                            all RSP/RP's are supported and in IOS XR RUN state
                            all FC's are supported and in OK state
                            all supported FAN/PEM's are in READY state
                            all supported LC's are in IOS XR RUN state
                            all MPA's are supported and in OK state

        default requirements (with override=False):
                            all RSP/RP's are supported and in IOS XR RUN state
                            all FC's are supported and in OK state
                            all FAN/PEM's are supported and in READY state
                            all supported LC's are in IOS XR RUN state
                            all MPA's are supported and in OK state

        :param inventory: the result for parsing the output of 'admin show platform'
        :param supported_hw: stores info about which card types are supported in eXR for RSP/RP/FAN/PEM/FC/LC
        :param override: override the requirement to check FAN/PEM hardware types
        :return: Errors out if a requirement is not met.
                If all requirements are met, returns a dictionary used later in other plugin(s) for monitoring
                the success of FPD upgrade. Keys are node names parsed from 'admin show platform'.
                 Values are integers. Value can either be 0 or 1. value 1 means that this node is supported in
                 eXR and is in valid state for migration. value 0 means that either this node is a PEM/FAN, or
                 this node is not supported in eXR.
        """
        fpd_relevant_nodes = {}

        for i in xrange(0, len(inventory)):
            node, entry = inventory[i]

            if node in fpd_relevant_nodes:
                continue

            if self.rp_pattern.match(node):
                rp_or_rsp = self._check_if_supported_and_in_valid_state(node, entry,
                                                                        supported_hw.get("RP"), "IOS XR RUN")

                fpd_relevant_nodes[node] = rp_or_rsp

            elif self.fc_pattern.match(node):

                fc = self._check_if_supported_and_in_valid_state(node, entry,
                                                                 supported_hw.get("FC"), "OK")
                fpd_relevant_nodes[node] = fc

            elif self.lc_pattern.match(node):

                lc = self._check_if_supported_and_in_valid_state(node, entry, supported_hw.get("LC"),
                                                                 "IOS XR RUN", mandatory=False)
                for j in xrange(i + 1, len(inventory)):
                    next_node, next_entry = inventory[j]
                    if "MPA" in next_entry["type"]:
                        mpa = self._check_if_supported_and_in_valid_state(next_node, next_entry,
                                                                          supported_hw.get("MPA"), "OK")
                        fpd_relevant_nodes[next_node] = mpa

                fpd_relevant_nodes[node] = lc

            elif self.fan_pattern.match(node):
                self._check_if_supported_and_in_valid_state(node, entry,
                                                            supported_hw.get("FAN"),
                                                            "READY", mandatory=not override)
                fpd_relevant_nodes[node] = 0
            elif self.pem_pattern.match(node):
                    self._check_if_supported_and_in_valid_state(node, entry,
                                                                supported_hw.get("PEM"),
                                                                "READY", mandatory=not override)
                    fpd_relevant_nodes[node] = 0
            else:
                fpd_relevant_nodes[node] = 1

        return fpd_relevant_nodes

    def _check_if_supported_and_in_valid_state(self, node_name, value,
                                               supported_type_list, operational_state, mandatory=True):
        """
        Check if a card (RSP/RP/FAN/PEM/FC/MPA) is supported and in valid state.
        :param node_name: the name under "Node" column in output of CLI "show platform". i.e., "0/RSP0/CPU0"
        :param value: the inventory value for nodes - through parsing output of "show platform"
        :param supported_type_list: the list of card types/pids that are supported for migration
        :param operational_state: the state that this node can be in in order to qualify for migration
        :param mandatory: if mandatory is True, if this node matches the card_pattern, it must be supported card
                          type in order to qualify for migration. If it's False, it's not necessary that the card
                          type is supported, but if it is supported, its state must be in the operational state
        :return: 1 if it's confirmed that the node is supported and in the operational state for migration.

                 0 if it's not mandatory that this node has supported card type, and it's confirmed that it does
                    not have supported card type, but it is in the operational state for migration.

                 errors out if this node is in either of the situations below:
                    1. It is mandatory for this node to have supported card type, but the card type of this node is NOT
                        supported for migration.
                    2. This node is supported for migration, but it is not in the operational state for migration.
        """
        supported = False
        if supported_type_list is None:
            self.ctx.error("The supported hardware list is missing information.")
        for supported_type in supported_type_list:
            if supported_type in value['type']:
                supported = True
                break
        if mandatory and not supported:
            self.ctx.error("The card type for {} is not supported for migration to ASR9K-X64.".format(node_name) +
                           " Please check the user manual under 'Help' on CSM Server for list of " +
                           "supported hardware for ASR9K-X64.")

        if supported and value['state'] != operational_state:
            self.ctx.error("{} is supported in ASR9K-X64, but it's in {}".format(node_name, value['state']) +
                           " state. Valid operational state for migration: {}".format(operational_state))
        if supported:
            return 1
        return 0

    def run(self):

        software_version_tuple = self.ctx.load_job_data('hardware_audit_version')
        if software_version_tuple:
            software_version = software_version_tuple[0]
            self.ctx.info("Hardware audit for software release version " + str(software_version))
        else:
            self.ctx.error("No software version selected.")

        override_hw_req_tuple = self.ctx.load_job_data('override_hw_req')
        if override_hw_req_tuple and override_hw_req_tuple[0] == "1":
            override_hw_req = True
            log_and_post_status(self.ctx,
                                "Running hardware audit to check minimal requirements for card types and states.")
        else:
            override_hw_req = False
            log_and_post_status(self.ctx, "Running hardware audit on all nodes.")

        with open(SUPPORTED_HW_JSON) as supported_hw_file:
            supported_hw = json.load(supported_hw_file)

        if not supported_hw.get(software_version):
            self.ctx.error("No hardware support information available for release {}.".format(software_version))

        # show platform can take more than 1 minute after router reload. Issue No. 47
        output = self.ctx.send("admin show platform", timeout=600)
        inventory = parse_admin_show_platform(output)

        log_and_post_status(self.ctx, "Check if cards on device are supported for migration.")
        fpd_relevant_nodes = self._check_if_hw_supported_and_in_valid_state(inventory,
                                                                            supported_hw[software_version],
                                                                            override_hw_req)
        self.ctx.save_job_data("fpd_relevant_nodes", fpd_relevant_nodes)

        log_and_post_status(self.ctx, "Hardware audit completed successfully.")

        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)

        return True



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import re
from csmpe.plugins import CSMPlugin
from install import install_add_remove
from csmpe.core_plugins.csm_get_inventory.ios.plugin import get_package, get_inventory


class Plugin(CSMPlugin):
    """This plugin adds packages from repository to the device."""
    name = "Install Add Plugin"
    platforms = {'ASR900'}
    phases = {'Add'}
    os = {'IOS'}

    def run(self):
        server_repository_url = self.ctx.server_repository_url

        if server_repository_url is None:
            self.ctx.error("No repository provided")
            return

        packages = self.ctx.software_packages
        if packages is None:
            self.ctx.error("No package list provided")
            return

        self.ctx.info("Add Package(s) Pending")
        self.ctx.post_status("Add Package(s) Pending")

        for package in packages:

            output = self.ctx.send('dir flash:' + package)
            m = re.search('No such file', output)

            if not m:
                self.ctx.info("No action: {} exists in flash:".format(package))
                continue

            cmd = "copy {}/{} flash:".format(server_repository_url, package)

            install_add_remove(self.ctx, cmd)

        self.ctx.info("Package(s) Added Successfully")

        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================


# from package_lib import SoftwarePackage
from csmpe.plugins import CSMPlugin
from install import install_activate_reload
from install import install_activate_write_memory
from csmpe.core_plugins.csm_get_inventory.ios.plugin import get_package, get_inventory
from csmpe.core_plugins.csm_install_operations.utils import update_device_info_udi


class Plugin(CSMPlugin):
    """This plugin Activates packages on the device."""
    name = "Install Activate Plugin"
    platforms = {'ASR900'}
    phases = {'Activate'}
    os = {'IOS'}

    def run(self):
        """
        Performs install activate operation
        """

        self.ctx.info("Hardware platform: {}".format(self.ctx._connection.platform))
        self.ctx.info("OS Version: {}".format(self.ctx._connection.os_version))

        try:
            packages = self.ctx.software_packages
        except AttributeError:
            self.ctx.warning("No package list provided. Skipping calculation of required free bootflash memory.")
            return

        pkg = ''.join(packages)

        if not pkg.startswith("asr901"):
            self.ctx.error('Unsupported image type for {} while Activate package = '
                           '{}'.format(self.ctx._connection.platform, pkg))

        self.ctx.info("Activate package(s) pending")
        self.ctx.post_status("Activate Package(s) Pending")

        # prompt = self.ctx._connection.hostname

        # configurations
        cmd = "configure terminal"
        self.ctx.send(cmd)
        cmd = "config-register 0x2102"
        self.ctx.send(cmd)

        cmd = "no boot system"
        self.ctx.send(cmd)
        cmd = "boot system flash:" + pkg
        self.ctx.send(cmd)

        self.ctx.send('end')

        cmd = "write memory"
        install_activate_write_memory(self.ctx, cmd, self.ctx._connection.hostname)
        # self.ctx.send(cmd, timeout=300, wait_for_string=prompt)

        # Start activation

        install_activate_reload(self.ctx)

        self.ctx.info("Activate package done")

        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)
        update_device_info_udi(self.ctx)

        # Verify the Image type

        output = self.ctx.send('show version | include ^System image')
        if pkg not in output:
            self.ctx.error('System image {} not found in show version: {}'.format(pkg, output))

        self.ctx.info('Activate image {} is successful'.format(pkg))



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================
import re
import time

plugin_ctx = None


install_error_pattern = re.compile("Error:    (.*)$", re.MULTILINE)


def log_install_errors(ctx, output):
    """
    Print warning for Error:

    :param ctx:
    :param output:
    :return: nothing
    """
    errors = re.findall(install_error_pattern, output)
    for line in errors:
        ctx.warning(line)


def remove_exist_image(ctx, package):
    """
    Remove the existing packages

    :param ctx
    :param package
    :return: True or False
    """

    output = ctx.send('dir ' + package)
    m = re.search('No such file', output)

    if m:
        return True
    else:
        cmd = "del /force {}".format(package)
        ctx.send(cmd)
        ctx.info("Removing files : {}".format(package))

    output = ctx.send(cmd)
    m = re.search('No such file', output)
    if m:
        return True
    else:
        return False


def send_newline(fsm_ctx):
    fsm_ctx.ctrl.sendline('\r\n')
    fsm_ctx.ctrl.sendline('\r\n')
    fsm_ctx.ctrl.sendline('\r\n')
    return True


def issu_error_state(fsm_ctx):
    plugin_ctx.warning("Error in ISSU. Please see session.log for details")
    return False


def wait_for_reload(ctx):
    """
     Wait for system to come up with max timeout as 25 Minutes

    """
    begin = time.time()
    ctx.disconnect()
    ctx.post_status("Waiting for device boot to reconnect")
    ctx.info("Waiting for device boot to reconnect")
    time.sleep(1500)   # 25 * 60 = 1500
    ctx.reconnect(force_discovery=True)   # default max_timeout=360
    ctx.info("Boot process finished")
    ctx.info("Device connected successfully")

    timeout = 3600
    poll_time = 30
    time_waited = 0

    ctx.info("Waiting for the device to come up")
    ctx.post_status("Waiting for the device to come up")
    time.sleep(30)

    output = None

    ncnt = 0
    while 1:

        ncnt += 1
        if ncnt > 20:
            break

        # Wait till all nodes are in XR run state
        time_waited += poll_time
        if time_waited >= timeout:
            break

        time.sleep(poll_time)

        output = ctx.send('show version | include ^System image')

        ctx.info("output = {}".format(output))

        m = re.search('(asr.*\.bin)', output)
        if m:
            ctx.info("The device is in the desired state")
            elapsed = time.time() - begin
            ctx.info("Overall outage time: {} minute(s) {:.0f} second(s)".format(elapsed // 60, elapsed % 60))
            return True

    # Some nodes did not come to run state
    ctx.error("Not all nodes have came up: {}".format(output))
    # this will never be executed
    return False


def install_activate_write_memory(ctx, cmd, hostname):
    """

    PAN-5201-ASR903#write memory
    Building configuration...
    [OK]
    PAN-5201-ASR903#

    PAN-5201-ASR903#write memory
    Warning: Attempting to overwrite an NVRAM configuration previously written
    by a different version of the system image.
    Overwrite the previous NVRAM configuration?[confirm]

    """
    global plugin_ctx
    plugin_ctx = ctx

    # Seeing this message without the reboot prompt indicates a non-reload situation
    Build_config = re.compile("\[OK\]")

    Overwrite_warning = re.compile("Overwrite the previous NVRAM configuration\?\[confirm\]")

    Host_prompt = re.compile(hostname)

    events = [Host_prompt, Overwrite_warning, Build_config]
    transitions = [
        (Overwrite_warning, [0], 1, send_newline, 1200),
        (Build_config, [0, 1], 2, None, 1200),
        (Host_prompt, [0, 1, 2], -1, None, 1200),
    ]

    if not ctx.run_fsm("write memory", cmd, events, transitions, timeout=1200):
        ctx.error("Failed: {}".format(cmd))


def install_add_remove(ctx, cmd):
    """
    Execute the copy command

    :param ctx
    :param cmd
    :return: nothing
    """
    message = "Waiting the operation to continue"
    ctx.info(message)
    ctx.post_status(message)

    ctx.send(cmd, wait_for_string="Destination filename")
    output = ctx.send("\r\n\r\n\r\n", timeout=3600)

    result = re.search("\d+ bytes copied in .* secs", output)

    if result:
        ctx.info("Command {} finished successfully".format(cmd))
        return
    else:
        log_install_errors(ctx, output)
        ctx.error("Command {} failed".format(cmd))


def install_activate_reload(ctx):
    """
    Reload the router

    :param ctx
    :return: nothing
    """
    message = "Waiting the {} operation to continue".format('reload')
    ctx.info(message)
    ctx.post_status(message)

    if not ctx.reload(reload_timeout=1200, no_reload_cmd=True):
        ctx.error("Encountered error when attempting to reload device.")

    success = wait_for_reload(ctx)

    if not success:
        ctx.error("Reload or boot failure")
        return

    ctx.info("Operation reload finished successfully")
    return



# =============================================================================
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================
import re


def parse_show_platform(ctx, output):
    """
    :param output: output from 'show platform'
    :return: dictionary of nodes

    Platform: ASR9K-X64
    Node              Type                       State             Config state
    --------------------------------------------------------------------------------
    0/RSP0/CPU0       A9K-RSP880-SE(Active)      IOS XR RUN        NSHUT
    0/RSP1/CPU0       A9K-RSP880-SE(Standby)     IOS XR RUN        NSHUT
    0/FT0             ASR-9904-FAN               OPERATIONAL       NSHUT
    0/0/CPU0          A9K-8X100GE-L-SE           IOS XR RUN        NSHUT
    0/PT0

    Platform: ASR9K-X64

    RP/0/RSP0/CPU0:vkg3#admin show platform
    Tue Apr 11 10:31:41.051 PST
    Node            Type                      State            Config State
    -----------------------------------------------------------------------------
    0/RSP0/CPU0     A9K-RSP880-SE(Active)     IOS XR RUN       PWR,NSHUT,MON
    0/RSP1/CPU0     A9K-RSP880-SE(Standby)    IOS XR RUN       PWR,NSHUT,MON
    0/FT0/SP        ASR-9904-FAN              READY
    0/0/CPU0        A9K-8X100GE-SE            IOS XR RUN       PWR,NSHUT,MON
    0/PS0/M0/SP     PWR-3KW-AC-V2             READY            PWR,NSHUT,MON
    0/PS0/M1/SP     PWR-3KW-AC-V2             READY            PWR,NSHUT,MON
    0/PS0/M2/SP     PWR-3KW-AC-V2             READY            PWR,NSHUT,MON
    0/PS0/M3/SP     PWR-3KW-AC-V2             READY            PWR,NSHUT,MON

    Platform: NCS6K
    RP/0/RP0/CPU0:ios#show platform

    Node              Type                       State             Config state
    --------------------------------------------------------------------------------
    0/0/CPU0          NC6-10X100G-M-K            IOS XR RUN        NSHUT
    0/0/NPU0          Slice                      UP
    0/0/NPU1          Slice                      UP
    0/0/NPU2          Slice                      UP
    0/0/NPU3          Slice                      UP
    0/0/NPU4          Slice                      UP
    0/5/CPU0          NC6-60X10GE-L-S            IOS XR RUN        NSHUT
    0/5/NPU0          Slice                      UP
    0/5/NPU1          Slice                      UP
    0/5/NPU2          Slice                      UP
    0/5/NPU3          Slice                      UP
    0/7/CPU0          NC6-20X100GE-M-C           IOS XR RUN        NSHUT
    0/7/NPU0          Slice                      UP
    0/7/NPU1          Slice                      UP
    0/7/NPU2          Slice                      UP
    0/7/NPU3          Slice                      UP
    0/7/NPU4          Slice                      UP
    0/RP0/CPU0        NC6-RP(Active)             IOS XR RUN        NSHUT
    0/RP1/CPU0        NC6-RP(Standby)            IOS XR RUN        NSHUT
    0/FC5             NC6-FC2-U                  OPERATIONAL       NSHUT
    0/CI0             NCS-CRFT                   OPERATIONAL       NSHUT
    0/FT0             NC6-FANTRAY-2              OPERATIONAL       NSHUT
    0/FT1             NC6-FANTRAY-2              OPERATIONAL       NSHUT
    0/PT0             NCS-AC-PWRTRAY             OPERATIONAL       NSHUT
    0/PT1             NCS-AC-PWRTRAY             OPERATIONAL       NSHUT
    0/PT2             NCS-AC-PWRTRAY             OPERATIONAL       NSHUT
    0/PT3             NCS-AC-PWRTRAY             OPERATIONAL       NSHUT
    0/PT4             NCS-AC-PWRTRAY             OPERATIONAL       NSHUT
    0/PT5             NCS-AC-PWRTRAY             OPERATIONAL       NSHUT

    Platform: NCS5500
    RP/0/RP0/CPU0:att-1#show platform

    Node              Type                       State             Config state
    --------------------------------------------------------------------------------
    0/1/CPU0          NC55-24X100G-SE            IOS XR RUN        NSHUT
    0/1/NPU0          Slice                      UP
    0/1/NPU1          Slice                      UP
    0/1/NPU2          Slice                      UP
    0/1/NPU3          Slice                      UP
    0/RP0/CPU0        NC55-RP(Active)            IOS XR RUN        NSHUT
    0/RP1/CPU0        NC55-RP(Standby)           IOS XR RUN        NSHUT
    0/FC1             UNKNOWN                    UNKNOWN
    0/FC5             NC55-5508-FC               OPERATIONAL       NSHUT
    0/FT0             NC55-5508-FAN              OPERATIONAL       NSHUT
    0/FT2             NC55-5508-FAN              OPERATIONAL       NSHUT
    0/SC0             UNKNOWN                    UNKNOWN
    0/SC1             NC55-SC                    OPERATIONAL       NSHUT

    Platform: NCS5K
    RP/0/RP0/CPU0:montreal#show platform


    Node              Type                       State             Config state

    --------------------------------------------------------------------------------

    0/RP0/CPU0        NCS-5001(Active)           IOS XR RUN        NSHUT

    0/FT0             NCS-5001-FN-BK             OPERATIONAL       NSHUT

    0/FT1             NCS-5001-FN-BK             OPERATIONAL       NSHUT

    Platform: NCS4K
    RP/0/RP0:DC2-4K#show platform

    Node              Type                       State             Config state
    --------------------------------------------------------------------------------
    0/0/CPU0          NCS4K-2H-W                 IOS XR RUN        NSHUT
    0/1/CPU0          NCS4K-2H10T-OP-KS          IOS XR RUN        NSHUT
    0/4/CPU0          NCS4K-2H-O-K               IOS XR RUN        NSHUT
    0/5/CPU0          NCS4K-24LR-O-S             IOS XR RUN        NSHUT
    0/10/CPU0         NCS4K-20T-O-S              IOS XR RUN        NSHUT
    0/11/CPU0         NCS4K-2H-O-K               IOS XR RUN        NSHUT
    0/13/CPU0         NCS4K-2H-O-K               IOS XR RUN        NSHUT
    0/15/CPU0         NCS4K-24LR-O-S             POWERED_OFF       NSHUT
    0/RP0/CPU0        NCS4K-RP                   IOS XR RUN        NSHUT
    0/RP1/CPU0        NCS4K-RP                   IOS XR RUN        NSHUT
    0/FC0             NCS4016-FC2-M              OPERATIONAL       NSHUT
    0/FC1             NCS4016-FC2-M              OPERATIONAL       NSHUT
    0/FC2             NCS4016-FC2-M              OPERATIONAL       NSHUT
    0/FC3             NCS4016-FC2-M              OPERATIONAL       NSHUT
    0/CI0             NCS4K-CRAFT                OPERATIONAL       NSHUT
    0/FT0             NCS4K-FTA                  OPERATIONAL       NSHUT
    0/FT1             NCS4K-FTA                  OPERATIONAL       NSHUT
    0/PT0             NCS4K-DC-PEM               OPERATIONAL       NSHUT
    0/PT1             NCS4K-DC-PEM               OPERATIONAL       NSHUT
    0/EC0             NCS4K-ECU                  OPERATIONAL       NSHUT

    Platform NCS1K:
    RP/0/RP0/CPU0:ios#show platform

    Node              Type                       State             Config state
    --------------------------------------------------------------------------------
    0/RP0/CPU0        NCS1K-CNTLR(Active)        IOS XR RUN        NSHUT

    Platform: XRv9K
    RP/0/RP0/CPU0:XRv9K#show platform
    Tue Apr 11 23:55:51.866 UTC
    Node              Type                       State             Config state
    --------------------------------------------------------------------------------
    0/RP0/CPU0        R-IOSXRV9000-RP(Active)    IOS XR RUN        NSHUT
    """
    inventory = {}
    sl = ['node', 'type', 'state', 'config state']
    dl = {}

    lines = output.split('\n')
    lines = [x for x in lines if x]
    for line in lines:
        line = line.strip()

        if line[0:4] == 'Node':
            line = line.lower()
            for s in sl:
                n = line.find(s)
                if n != -1:
                    dl[str(s)] = n
                else:
                    ctx.warning("unrecognized show platform header {}".format(line))
                    return None

        if line[0].isdigit():
            node = line[:dl["type"]]
            if not re.search('CPU\d+\s*$', node):
                continue

            node_type = line[dl['type']:dl['state']].strip()
            state = line[dl['state']:dl['config state']].strip()
            config_state = line[dl['config state']:].strip()

            entry = {
                'type': node_type,
                'state': state,
                'config_state': config_state
            }
            inventory[node] = entry

    return inventory



# coding=utf-8
# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================
from functools import partial
import itertools
import re
import time
from condoor import ConnectionError, CommandError
from csmpe.core_plugins.csm_node_status_check.exr.plugin_lib import parse_show_platform
from csmpe.core_plugins.csm_install_operations.actions import a_error

install_error_pattern = re.compile("Error:    (.*)$", re.MULTILINE)

plugin_ctx = None


def log_install_errors(ctx, output):
        errors = re.findall(install_error_pattern, output)
        for line in errors:
            ctx.warning(line)


def check_ncs6k_release(ctx):
    """
    Only release 5.2.5 and above are supported by the plugin.
    """
    if ctx.family == 'NCS6K':
        packages = ctx.software_packages
        if packages is not None:
            for package_name in packages:
                matches = re.findall("\d+\.\d+\.\d+", package_name)
                if matches:
                    if matches[0] < '5.2.5':
                        ctx.error('Abort: Software package earlier than release 5.2.5 for NCS6K is not supported.')


def check_ncs4k_release(ctx):
    """
    Only release 6.0.2 and above are supported by the plugin.
    """
    if ctx.family == 'NCS4K':
        packages = ctx.software_packages
        if packages is not None:
            for package_name in packages:
                matches = re.findall("\d+\.\d+\.\d+", package_name)
                if matches:
                    if matches[0] < '6.0.2':
                        ctx.error('Abort: Software package earlier than release 6.0.2 for NCS4K is not supported.')


def watch_operation(ctx, op_id=0):
    """
    Watch for the non-reload situation.  Upon issuing add/activate/commit/remove/deactivate, the install operation
    will be executed in the background.  The following message

    Install operation will continue in the background

    will be displayed.  After some time elapses, a successful or abort message will be displayed.

    The CLI command, 'show install request' is used in the loop to report the progress percentages.  Upon
    completion, 'show install request' returns 'No install operation in progress'.  The watch_operation will be
    done at that point.

    As an example,

    RP/0/RP0/CPU0:Deploy#install deactivate ncs6k-5.2.5.CSCuz65240-1.0.0
    May 31 20:14:14 Install operation 3 started by root:
      install deactivate pkg ncs6k-5.2.5.CSCuz65240-1.0.0
    May 31 20:14:14 Package list:
    May 31 20:14:14     ncs6k-5.2.5.CSCuz65240-1.0.0
    May 31 20:14:20 Install operation will continue in the background

    <--- Time Lapse --->

    RP/0/RP0/CPU0:Deploy#May 31 20:15:10 Install operation 3 finished successfully

    ------------------------------------------------------------------------------------------------------------
    RP/0/RP0/CPU0:Deploy#show install request
    The install operation 17 is 30% complete

    or

    RP/0/RP0/CPU0:Deploy#show install request
    No install operation in progress

    When install is completed, the following message will be displayed

    ------------------------------------------------------------------------------------------------------------
    If the install operation is successful, the following message will be displayed.

    RP/0/RP0/CPU0:Deploy#May 24 22:25:43 Install operation 17 finished successfully
    """
    no_install = r"No install operation in progress"
    # In ASR9K eXR, the output to show install request may be "The install prepare operation 9 is 40% complete"
    # or "The install service operation 9 is 40% complete" or "The install add operation 9 is 40% complete" and etc.
    op_progress = r"The install \w*?\s?operation {} is (\d+)% complete".format(op_id)
    success = "Install operation {} finished successfully".format(op_id)

    cmd_show_install_request = "show install request"
    ctx.info("Watching the operation {} to complete".format(op_id))

    propeller = itertools.cycle(["|", "/", "-", "\\", "|", "/", "-", "\\"])

    last_status = None
    finish = False
    time_tried = 0
    while not finish:
        try:
            try:
                # this is to catch the successful operation as soon as possible
                ctx.send("", wait_for_string=success, timeout=20)
                finish = True
            except ctx.CommandTimeoutError:
                pass

            message = ""
            output = ctx.send(cmd_show_install_request, timeout=300)
            if op_id in output:
                result = re.search(op_progress, output)
                if result:
                    status = result.group(0)
                    message = "{} {}".format(propeller.next(), status)

                if message != last_status:
                    ctx.post_status(message)
                    last_status = message
        except (ConnectionError, ctx.CommandTimeoutError) as e:
            if time_tried > 2:
                raise e

            time_tried += 1
            ctx.disconnect()
            time.sleep(60)
            ctx.reconnect(force_discovery=True)

        if no_install in output:
            break

    report_install_status(ctx, op_id)


def validate_node_state(inventory):
    valid_state = [
        'IOS XR RUN',
        'PRESENT',
        'READY',
        'FAILED',
        'OK',
        'DISABLED',
        'UNPOWERED',
        'POWERED_OFF',
        'SW_INACTIVE',
        'ADMIN DOWN',
        'OPERATIONAL',
        'NOT ALLOW ONLIN',  # This is not spelling error
    ]

    for key, value in inventory.items():
        if 'CPU' in key:
            if value['state'] not in valid_state:
                break
    else:
        return True

    return False


def wait_for_reload(ctx):
    """
     Wait for system to come up with max timeout as 25 Minutes

    """
    begin = time.time()
    if not ctx.is_console:
        ctx.disconnect()
        ctx.post_status("Waiting for device boot to reconnect")
        ctx.info("Waiting for device boot to reconnect")
        time.sleep(60)
        ctx.reconnect(max_timeout=3600, force_discovery=True)  # 60 * 60 = 3600

    else:
        ctx.info("Keeping console connected")
        ctx.post_status("Boot process started")
        ctx.info("Boot process started")
        if not ctx.reload(reload_timeout=1500, no_reload_cmd=True):
            ctx.error("Encountered error when attempting to reload device.")
        ctx.info("Boot process finished")

    ctx.info("Device connected successfully")

    timeout = 3600
    poll_time = 30
    time_waited = 0
    xr_run = "IOS XR RUN"

    cmd = "show platform"
    ctx.info("Waiting for all nodes to come up")
    ctx.post_status("Waiting for all nodes to come up")

    time.sleep(100)

    while 1:
        # Wait till all nodes are in XR run state
        time_waited += poll_time
        if time_waited >= timeout:
            break

        time.sleep(poll_time)

        # show platform can take more than 1 minute after router reload. Issue No. 47
        output = ctx.send(cmd, timeout=600)
        if xr_run in output:
            inventory = parse_show_platform(ctx, output)
            if validate_node_state(inventory):
                ctx.info("All nodes in desired state")
                elapsed = time.time() - begin
                ctx.info("Overall outage time: {} minute(s) {:.0f} second(s)".format(elapsed // 60, elapsed % 60))
                return True

    # Some nodes did not come to run state
    ctx.error("Not all nodes have came up: {}".format(output))
    # this will never be executed
    return False


def observe_install_add_remove(ctx, output, has_tar=False):
    """
    Success Condition:
    ADD:
    install add source tftp://223.255.254.254/auto/tftpboot-users/alextang/ ncs6k-mpls.pkg-6.1.0.07I.DT_IMAGE
    May 24 18:54:12 Install operation will continue in the background
    RP/0/RP0/CPU0:Deploy#May 24 18:54:30 Install operation 12 finished successfully

    REMOVE:
    RP/0/RP0/CPU0:Deploy#install remove ncs6k-5.2.5.47I.CSCux97367-0.0.15.i
    May 23 21:20:28 Install operation 2 started by root:
      install remove ncs6k-5.2.5.47I.CSCux97367-0.0.15.i
    May 23 21:20:28 Package list:
    May 23 21:20:28     ncs6k-5.2.5.47I.CSCux97367-0.0.15.i
    May 23 21:20:29 Install operation will continue in the background

    RP/0/RP0/CPU0:Deploy#May 23 21:20:29 Install operation 2 finished successfully

    Failed Condition:
    RP/0/RSP0/CPU0:CORFU#install remove ncs6k-5.2.5.47I.CSCux97367-0.0.15.i
    Mon May 23 22:57:45.078 UTC
    May 23 22:57:46 Install operation 28 started by iox:
      install remove ncs6k-5.2.5.47I.CSCux97367-0.0.15.i
    May 23 22:57:46 Package list:
    May 23 22:57:46     ncs6k-5.2.5.47I.CSCux97367-0.0.15.i
    May 23 22:57:47 Install operation will continue in the background

    RP/0/RSP0/CPU0:CORFU#May 23 22:57:48 Install operation 28 aborted
    """
    result = re.search('Install operation (\d+)', output)
    if result:
        op_id = result.group(1)
    else:
        log_install_errors(ctx, output)
        ctx.error("Operation failed")
        return  # for sake of clarity

    op_success = "Install operation will continue in the background"

    if op_success in output:
        watch_operation(ctx, op_id)
        if has_tar is True:
            ctx.set_operation_id(ctx.software_packages, op_id)
            ctx.info("The operation {} stored".format(op_id))
    else:
        log_install_errors(ctx, output)
        ctx.error("Operation {} failed".format(op_id))


def get_op_id(output):
    """
    :param output: Output from the install command
    :return: the operational ID
    """
    result = re.search('Install operation (\d+)', output)
    if result:
        return result.group(1)
    return -1


def report_install_status(ctx, op_id):
    """
    :param ctx: CSM Context object
    :param op_id: operational ID
    Peeks into the install log to see if the install operation is successful or not
    """
    failed_oper = r'Install operation {} aborted'.format(op_id)
    output = ctx.send("show install log {} detail".format(op_id))
    if re.search(failed_oper, output):
        log_install_errors(ctx, output)
        ctx.error("Operation {} failed".format(op_id))

    ctx.info("Operation {} finished successfully".format(op_id))


def handle_aborted(fsm_ctx):
    """
    :param ctx: FSM Context
    :return: True if successful other False
    """
    global plugin_ctx

    report_install_status(ctx=plugin_ctx, op_id=get_op_id(fsm_ctx.ctrl.before))

    # Indicates the failure
    return False


def handle_non_reload_activate_deactivate(fsm_ctx):
    """
    :param ctx: FSM Context
    :return: True if successful other False
    """
    global plugin_ctx

    op_id = get_op_id(fsm_ctx.ctrl.before)
    if op_id == -1:
        return False

    watch_operation(plugin_ctx, op_id)

    return True


def handle_reload_activate_deactivate(fsm_ctx):
    """
    :param ctx: FSM Context
    :return: True if successful other False
    """
    global plugin_ctx

    op_id = get_op_id(fsm_ctx.ctrl.before)
    if op_id == -1:
        return False

    try:
        watch_operation(plugin_ctx, op_id)
    except plugin_ctx.CommandTimeoutError:
        # The device already started the reload
        pass

    success = wait_for_reload(plugin_ctx)
    if not success:
        plugin_ctx.error("Reload or boot failure")
        return

    plugin_ctx.info("Operation {} finished successfully".format(op_id))

    return True


def no_impact_warning(fsm_ctx):
    plugin_ctx.warning("This was a NO IMPACT OPERATION. Packages are already active on device.")
    return True


def handle_not_start(fsm_ctx):
    """
    :param ctx: FSM Context
    :return: False
    """
    global plugin_ctx

    plugin_ctx.error("Could not start this install operation because an install operation is still in progress")
    return False


def install_activate_deactivate(ctx, cmd):
    """
    Abort Situation:
    RP/0/RP0/CPU0:Deploy#install activate ncs6k-5.2.5.CSCuz65240-1.0.0

    Jun 02 20:19:31 Install operation 8 started by root:
      install activate pkg ncs6k-5.2.5.CSCuz65240-1.0.0
    Jun 02 20:19:31 Package list:
    Jun 02 20:19:31     ncs6k-5.2.5.CSCuz65240-1.0.0
    Jun 02 20:19:31     ncs6k-5.2.5.47I.CSCuy47880-0.0.4.i
    Jun 02 20:19:31     ncs6k-5.2.5.CSCux82987-1.0.0
    Jun 02 20:19:38 Install operation 8 aborted

    ------------------------------------------------------------------------------------------------------------
    Non-Reload Situation:

    RP/0/RP0/CPU0:Deploy#install deactivate ncs6k-5.2.5.CSCuz65240-1.0.0
    May 31 20:14:14 Install operation 3 started by root:
      install deactivate pkg ncs6k-5.2.5.CSCuz65240-1.0.0
    May 31 20:14:14 Package list:
    May 31 20:14:14     ncs6k-5.2.5.CSCuz65240-1.0.0
    May 31 20:14:20 Install operation will continue in the background (may or may not be displayed)

    <--- Time Lapses --->

    RP/0/RP0/CPU0:Deploy#May 31 20:15:10 Install operation 3 finished successfully

    ------------------------------------------------------------------------------------------------------------
    Reload Situation 1:

    RP/0/RP0/CPU0:Deploy#install activate ncs6k-5.2.5.CSCux82987-1.0.0
    May 31 20:17:08 Install operation 4 started by root:
      install activate pkg ncs6k-5.2.5.CSCux82987-1.0.0
    May 31 20:17:08 Package list:
    May 31 20:17:08     ncs6k-5.2.5.CSCux82987-1.0.0

    <--- Time Lapses --->

    This install operation will reboot the sdr, continue?
     [yes/no]:[yes] <Hit Enter>
    May 31 20:17:47 Install operation will continue in the background

    <--- Time Lapses --->

    RP/0/RP0/CPU0:Deploy#May 31 20:18:44 Install operation 4 finished successfully

    <--- Router Starts Reloading --->

    Connection closed by foreign host.

    Reload Situation 2:

    RP/0/RSP0/CPU0:ios#install activate id 25
    Jun 09 15:49:15 Install operation 27 started by root:
    install activate id 25
    Jun 09 15:49:15 Package list:
    Jun 09 15:49:15     asr9k-sysadmin-system-6.1.1.17-r61116I.CSCcv44444.x86_64
    Jun 09 15:49:15     asr9k-os-supp-64-3.1.0.1-r61116I.CSCxr90021.x86_64
    Jun 09 15:49:15     asr9k-base-64-4.0.0.2-r61116I.CSCxr90014.x86_64
    Jun 09 15:49:15     asr9k-sysadmin-topo-6.1.1.17-r61116I.CSCcv55555.x86_64

    This install operation will reload the sdr, continue?
    [yes/no]:[yes] <Hit Enter>
    Jun 09 15:49:26 Install operation will continue in the background
    RP/0/RSP0/CPU0:ios#

    <--- Time Lapses --->

    RP/0/RSP0/CPU0:ios#Jun 09 15:53:51 Install operation 27 finished successfully

    """
    global plugin_ctx
    plugin_ctx = ctx

    ABORTED = re.compile("aborted")

    # REBOOT_PROMP for eXR reload response:
    # This install operation will reload the sdr, continue?
    #  [yes/no]:[yes]
    # REBOOT_PROMP for NCS5K reload response:
    # This install operation will reload the system, continue?
    #  [yes/no]:[yes]

    # Seeing this message without the reboot prompt indicates a non-reload situation
    CONTINUE_IN_BACKGROUND = re.compile("Install operation will continue in the background")
    REBOOT_PROMPT = re.compile("This install operation will (?:reboot|reload) the sdr, continue")
    RELOAD_PROMPT = re.compile("This install operation will reload the system, continue")
    RUN_PROMPT = re.compile("#")
    NO_IMPACT = re.compile("NO IMPACT OPERATION")
    ERROR = re.compile(re.escape("ERROR! there was an SU/ISSU done. please perform install commit before "
                                 "proceeding with any other prepare/activate/deactivate operation"))
    NOT_START = re.compile("Could not start this install operation")

    #                  0                    1              2           3          4         5          6
    events = [CONTINUE_IN_BACKGROUND, REBOOT_PROMPT, RELOAD_PROMPT, ABORTED, NO_IMPACT, RUN_PROMPT, ERROR, NOT_START]
    transitions = [
        (CONTINUE_IN_BACKGROUND, [0], -1, handle_non_reload_activate_deactivate, 300),
        (REBOOT_PROMPT, [0], -1, handle_reload_activate_deactivate, 300),
        (RELOAD_PROMPT, [0], -1, handle_reload_activate_deactivate, 300),
        (NO_IMPACT, [0], -1, no_impact_warning, 60),
        (RUN_PROMPT, [0], -1, handle_non_reload_activate_deactivate, 300),
        (ABORTED, [0], -1, handle_aborted, 300),
        (ERROR, [0], -1, partial(a_error, ctx), 0),
        (NOT_START, [0], -1, handle_not_start, 300),
    ]

    if not ctx.run_fsm("ACTIVATE-OR-DEACTIVATE", cmd, events, transitions, timeout=300):
        ctx.error("Failed: {}".format(cmd))


def send_admin_cmd(ctx, cmd):
    ctx.send("admin")
    output = ctx.send(cmd)
    ctx.send("exit")

    return output


def observe_install_remove_all(ctx, cmd, prompt):
    """
    Success Condition:
    RP/0/RSP0/CPU0:vkg3#install remove inactive all
    Mar 01 16:17:13 Install operation 15 started by root:
      install remove inactive all
    Mar 01 16:17:14 Install operation will continue in the background
    RP/0/RSP0/CPU0:vkg3#Mar 01 16:17:18 Install operation 15 finished successfully

    Failed Conditions:
    RP/0/RSP0/CPU0:vkg3#install remove inactive all
    Mar 01 21:31:22 Install operation 17 started by root:
      install remove inactive all
    Mar 01 21:31:24 Install operation will continue in the background
    RP/0/RSP0/CPU0:vkg3#Mar 01 21:31:25 Install operation 17 aborted

    RP/0/RSP0/CPU0:vkg3#install remove inactive all
    Could not start this install operation. Install operation 18 is still in progress
    """
    # no op_id is returned from XR for install remove inactive
    # need to figure out the last op_id first

    cmd_show_install_log_reverse = 'show install log reverse | utility egrep "Install operation [0-9]+ started"'
    output = ctx.send(cmd_show_install_log_reverse, timeout=300)

    if 'No install operation' in output:
        op_id = 0
    else:
        result = re.search('Install operation (\d+) started', output)
        if result:
            op_id = int(result.group(1))
        else:
            log_install_errors(ctx, output)
            ctx.error("Operation ID not found by show install log reverse")
            return

    # Expected Operation ID
    op_id += 1

    oper_err = "Install operation {} aborted".format(op_id)
    oper_success = "Install operation {} finished successfully".format(op_id)
    Error1 = re.compile("Could not start this install operation. Install operation")
    Error2 = re.compile(oper_err)
    Proceed_removing = re.compile(oper_success)
    Host_prompt = re.compile(prompt)

    events = [Host_prompt, Error1, Error2, Proceed_removing]
    transitions = [
        (Error1, [0], -1, CommandError("Another install command is currently in operation",
                                       ctx._connection.hostname), 1800),
        (Error2, [0], -1, CommandError("No packages can be removed", ctx._connection.hostname), 1800),
        (Proceed_removing, [0], -1, None, 1800),
        (Host_prompt, [0], -1, None, 1800),
    ]

    if not ctx.run_fsm("Remove Inactive All", cmd, events, transitions, timeout=1800):
        ctx.error("Failed: {}".format(cmd))

    message = "Waiting the operation to continue asynchronously"
    ctx.info(message)
    ctx.post_status(message)

    last_status = None
    no_install = r"No install operation in progress"
    op_progress = r"The install operation {} is (\d+)% complete".format(op_id)
    cmd_show_install_request = "show install request"
    propeller = itertools.cycle(["|", "/", "-", "\\", "|", "/", "-", "\\"])

    finish = False
    time_tried = 0
    op_id = str(op_id)
    while not finish:
        try:
            try:
                # this is to catch the successful operation as soon as possible
                ctx.send("", wait_for_string=oper_success, timeout=20)
                finish = True
            except ctx.CommandTimeoutError:
                pass

            message = ""
            output = ctx.send(cmd_show_install_request, timeout=300)
            if op_id in output:
                result = re.search(op_progress, output)
                if result:
                    status = result.group(0)
                    message = "{} {}".format(propeller.next(), status)

                if message != last_status:
                    ctx.post_status(message)
                    last_status = message
        except (ConnectionError, ctx.CommandTimeoutError) as e:
            if time_tried > 120:
                raise e

            time_tried += 1
            time.sleep(30)

        if no_install in output:
            break

    cmd_show_install_log = "show install log {} detail".format(op_id)
    output = ctx.send(cmd_show_install_log, timeout=600)
    ctx.info(output)

    if oper_success in output:
        message = "Remove All Inactive Package(s) Successfully"
        ctx.info(message)
        ctx.post_status(message)
    else:
        ctx.error("Remove All Inactive Package(s) failed")



# =============================================================================
# migrate_system.py - plugin for migrating classic XR to eXR/fleXR
#
# Copyright (c)  2013, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import re
import time

from csmpe.plugins import CSMPlugin
from migration_lib import wait_for_final_band, log_and_post_status, run_additional_custom_commands
from csmpe.core_plugins.csm_get_inventory.exr.plugin import get_package, get_inventory

TIMEOUT_FOR_COPY_CONFIG = 3600


class Plugin(CSMPlugin):
    """
    A plugin for loading configurations and upgrading FPD's
    after the system migrated to ASR9K IOS-XR 64 bit(eXR).
    If any FPD needs reloads after upgrade, the device
    will be reloaded after the upgrade.
    Console access is needed.
    """
    name = "Post-Migrate Plugin"
    platforms = {'ASR9K'}
    phases = {'Post-Migrate'}

    def _check_fpds_for_upgrade(self):
        """Check if any FPD's need upgrade, if so, upgrade all FPD's on all locations."""

        self.ctx.send("admin")

        fpdtable = self.ctx.send("show hw-module fpd")

        match = re.search("\d+/\w+.+\d+.\d+\s+[-\w]+\s+(NEED UPGD)", fpdtable)

        if match:
            total_num = len(re.findall("NEED UPGD", fpdtable)) + len(re.findall("CURRENT", fpdtable))
            if not self._upgrade_all_fpds(total_num):
                self.ctx.send("exit")
                self.ctx.error("FPD upgrade in eXR is not finished. Please check session.log.")
                return False
            else:
                return True

        self.ctx.send("exit")
        return True

    def _upgrade_all_fpds(self, num_fpds):
        """
        Upgrade all FPD's on all locations.
        If after all upgrade completes, some show that a reload is required to reflect the changes,
        the device will be reloaded.

        :param num_fpds: the number of FPD's that are in CURRENT and NEED UPGD states before upgrade.
        :return: True if upgraded successfully and reloaded(if necessary).
                 False if some FPD's did not upgrade successfully in 9600 seconds.
        """
        log_and_post_status(self.ctx, "Upgrading all FPD's.")
        self.ctx.send("upgrade hw-module location all fpd all")

        timeout = 9600
        poll_time = 30
        time_waited = 0

        time.sleep(60)
        while 1:
            # Wait till all FPDs finish upgrade
            time_waited += poll_time
            if time_waited >= timeout:
                break
            time.sleep(poll_time)
            output = self.ctx.send("show hw-module fpd")
            num_need_reload = len(re.findall("RLOAD REQ", output))
            if len(re.findall("CURRENT", output)) + num_need_reload >= num_fpds:
                if num_need_reload > 0:
                    log_and_post_status(self.ctx,
                                        "Finished upgrading FPD(s). Now reloading the device to complete the upgrade.")
                    self.ctx.send("exit")
                    return self._reload_all()
                self.ctx.send("exit")
                return True

        # Some FPDs didn't finish upgrade
        return False

    def _reload_all(self):
        """Reload the device with 1 hour maximum timeout"""
        if self.ctx.reload(reload_timeout=3600):
            return self._wait_for_reload()
        self.ctx.error("Encountered error when attempting to reload device.")

    def _wait_for_reload(self):
        """Wait for all nodes to come up with max timeout as 18 min"""
        # device.disconnect()
        # device.reconnect(max_timeout=300)
        log_and_post_status(self.ctx, "Waiting for all nodes to come to FINAL Band.")
        if wait_for_final_band(self.ctx):
            log_and_post_status(self.ctx, "All nodes are in FINAL Band.")
        else:
            log_and_post_status(self.ctx, "Warning: Not all nodes went to FINAL Band.")

        return True

    def run(self):

        log_and_post_status(self.ctx, "Waiting for all nodes to come to FINAL Band.")
        if not wait_for_final_band(self.ctx):
            self.ctx.warning("Warning: Not all nodes are in FINAL Band after 25 minutes.")

        log_and_post_status(self.ctx, "Capturing new IOS XR and Calvados configurations.")

        run_additional_custom_commands(self.ctx, {"show running-config", "admin show running-config"})

        self._check_fpds_for_upgrade()

        run_additional_custom_commands(self.ctx, {"show platform"})

        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)



#!/usr/bin/env python
# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

try:
    import click
except ImportError:
    print("Install click python package\n pip install click")
    exit()

import logging
import os
import textwrap
import urlparse

from csmpe.context import InstallContext
from csmpe.csm_pm import CSMPluginManager
from csmpe.csm_pm import install_phases

_PLATFORMS = ["ASR9K", "NCS4K", "NCS6K", "CRS", "ASR900"]
_OS = ["IOS", "XR", "eXR", "XE"]


def print_plugin_info(pm, detail=False, brief=False):
    for plugin, details in pm.plugins.items():
        platforms = ", ".join(details['platforms'])
        phases = ", ".join(details['phases']) if bool(details['phases']) else "Any"
        os = ", ".join(details['os']) if bool(details['os']) else "Any"
        if brief:
            click.echo("[{}] [{}] [{}] {}".format(platforms, phases, os, details['name']))
        else:
            click.echo("Name: {}".format(details['name']))
            click.echo("Platforms: {}".format(platforms))
            click.echo("Phases: {}".format(phases))
            click.echo("OS: {}".format(os))
            description = "Description: {}\n".format(details['description'])
            description = "\n".join(textwrap.wrap(description, 60))
            click.echo(description)

            if detail:
                click.echo("  UUID: {}".format(plugin))
                package_name = details['package_name']
                click.echo("  Package Name: {}".format(package_name))
                pkginfo = pm.get_package_metadata(package_name)
                click.echo("  Summary: {}".format(pkginfo.summary))
                click.echo("  Version: {}".format(pkginfo.version))
                click.echo("  Author: {}".format(pkginfo.author))
                click.echo("  Author Email: {}".format(pkginfo.author_email))
            click.echo()


def validate_phase(ctx, param, value):
    if value:
        if value.strip() not in install_phases:
            raise click.BadParameter("The supported plugin phases are: {}".format(", ".join(install_phases)))
    return value


class URL(click.ParamType):
    name = 'url'

    def convert(self, value, param, ctx):
        if not isinstance(value, tuple):
            parsed = urlparse.urlparse(value)
            if parsed.scheme not in ('telnet', 'ssh'):
                self.fail('invalid URL scheme (%s).  Only telnet and ssh URLs are '
                          'allowed' % parsed, param, ctx)
        return value


@click.group()
def cli():
    """This script allows maintaining and executing the plugins."""
    pass


@cli.command("list", help="List all the plugins available.", short_help="List plugins")
@click.option("--platform", type=click.Choice(_PLATFORMS),
              help="Supported platform.")
@click.option("--phase", type=click.Choice(install_phases),
              help="Supported phase.")
@click.option("--os", type=click.Choice(_OS),
              help="Supported OS.")
@click.option("--detail", is_flag=True,
              help="Display detailed information about installed plugins.")
@click.option("--brief", is_flag=True,
              help="Display brief information about installed plugins.")
def plugin_list(platform, phase, os, detail, brief):
    pm = CSMPluginManager(None, invoke_on_load=False)
    pm.set_phase_filter(phase)
    pm.set_platform_filter(platform)
    pm.set_os_filter(os)
    pm.load(invoke_on_load=False)

    click.echo("List of installed plugins:\n")
    if platform:
        click.echo("Plugins for platform: {}".format(platform))
    if phase:
        click.echo("Plugins for phase: {}".format(phase))
    if os:
        click.echo("Plugins for os: {}".format(os))

    print_plugin_info(pm, detail, brief)


@cli.command("run", help="Run specific plugin on the device.", short_help="Run plugin")
@click.option("--url", multiple=True, required=True, envvar='CSMPLUGIN_URLS', type=URL(),
              help='The connection url to the host (i.e. telnet://user:pass@hostname). '
                   'The --url option can be repeated to define multiple jumphost urls. '
                   'If no --url option provided the CSMPLUGIN_URLS environment variable is used.')
@click.option("--phase", required=False, type=click.Choice(install_phases),
              help="An install phase to run the plugin for.")
@click.option("--cmd", multiple=True, default=[],
              help='The command to be passed to the plugin in ')
@click.option("--log_dir", default="/tmp", type=click.Path(),
              help="An install phase to run the plugin for. If not path specified then default /tmp directory is used.")
@click.option("--package", default=[], multiple=True,
              help="Package for install operations. This package option can be repeated to provide multiple packages.")
@click.option("--repository_url", default=None,
              help="The package repository URL. (i.e. tftp://server/dir")
@click.argument("plugin_name", required=False, default=None)
def plugin_run(url, phase, cmd, log_dir, package, repository_url, plugin_name):

    ctx = InstallContext()
    ctx.hostname = "Hostname"
    ctx.host_urls = list(url)
    ctx.success = False

    ctx.requested_action = phase
    ctx.log_directory = log_dir
    session_filename = os.path.join(log_dir, "session.log")
    plugins_filename = os.path.join(log_dir, "plugins.log")
    condoor_filename = os.path.join(log_dir, "condoor.log")

    if os.path.exists(session_filename):
        os.remove(session_filename)
    if os.path.exists(plugins_filename):
        os.remove(plugins_filename)
    if os.path.exists(condoor_filename):
        os.remove(condoor_filename)

    ctx.log_level = logging.DEBUG
    ctx.software_packages = list(package)
    ctx.server_repository_url = repository_url

    if cmd:
        ctx.custom_commands = list(cmd)

    pm = CSMPluginManager(ctx)
    pm.set_name_filter(plugin_name)
    results = pm.dispatch("run")

    click.echo("\n Plugin execution finished.\n")
    click.echo("Log files dir: {}".format(log_dir))
    click.echo(" {} - device session log".format(session_filename))
    click.echo(" {} - plugin execution log".format(plugins_filename))
    click.echo(" {} - device connection debug log".format(condoor_filename))
    click.echo("Results: {}".format(" ".join(map(str, results))))


if __name__ == '__main__':
    cli()



# =============================================================================
# plugin
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import re
import string
from csmpe.plugins import CSMPlugin
from csmpe.core_plugins.csm_install_operations.ios_xe.utils import number_of_rsp


class Plugin(CSMPlugin):
    """
    This plugin checks if the node redundancy is sufficient to proceed the upgrade.
    """
    name = "Node Redundancy Check Plugin"
    platforms = {'ASR900'}
    phases = {'Pre-Upgrade', 'Post-Upgrade'}

    def run(self):
        """
        PAN-5201-ASR903#show redundancy states
        my state = 13 -ACTIVE
        peer state = 8  -STANDBY HOT
        Mode = Duplex
        Unit = Secondary
        Unit ID = 49

        Redundancy Mode (Operational) = sso
        Redundancy Mode (Configured)  = sso
        Redundancy State              = sso
        Maintenance Mode = Disabled
        Manual Swact = enabled
        Communications = Up

        client count = 107
        client_notification_TMR = 30000 milliseconds
        RF debug mask = 0x0
        """

        sub_platforms = ['ASR-903', 'ASR-907']
        rsp_count = 1

        if self.ctx._connection.platform not in sub_platforms:
            self.ctx.info("Node redundancy not supported by "
                          "{}".format(self.ctx._connection.platform))
            return

        rsp_count = number_of_rsp(self.ctx)
        if rsp_count != 2:
            self.ctx.info("Node redundancy not supported by "
                          " number of RSP: {}".format(rsp_count))
            return

        cmd = "show redundancy states"
        output = self.ctx.send(cmd)
        if not output:
            self.ctx.error("Show redundancy output is insufficient.")
            return

        sso_ready = 0

        lines = string.split(output, '\n')
        lines = [x for x in lines if x]

        for line in lines:
            m = re.search('my state = .* -(.*)', line)
            if m:
                state = m.group(1)
                if 'ACTIVE' in state:
                    sso_ready = sso_ready | 1
                    self.ctx.info('{}'.format(line.lstrip()))
                continue

            m = re.search('peer state = .* -(.*)', line)
            if m:
                state = m.group(1)
                if 'STANDBY HOT' in state:
                    sso_ready = sso_ready | 2
                    self.ctx.info('{}'.format(line.lstrip()))
                continue

            m = re.search('Redundancy Mode \(Operational\)\s+= (.*)', line)
            if m:
                state = m.group(1)
                if 'sso' in state:
                    sso_ready = sso_ready | 4
                    self.ctx.info('{}'.format(line.lstrip()))
                continue

            m = re.search('Redundancy Mode \(Configured\)\s+= (.*)', line)
            if m:
                state = m.group(1)
                if 'sso' in state:
                    sso_ready = sso_ready | 8
                    self.ctx.info('{}'.format(line.lstrip()))
                continue

            m = re.search('Redundancy State        \s+ = (.*)', line)
            if m:
                state = m.group(1)
                if 'sso' in state:
                    sso_ready = sso_ready | 16
                    self.ctx.info('{}'.format(line.lstrip()))
                continue

        self.ctx.info("sso_ready = {}".format(sso_ready))
        if sso_ready == 31:
            self.ctx.info("Router redundancy has reached SSO state.")
        else:
            self.ctx.warning("Router redundancy has not reached SSO state.")



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

from package_lib import SoftwarePackage
from csmpe.plugins import CSMPlugin
from install import install_activate_deactivate
from csmpe.core_plugins.csm_get_inventory.ios_xr.plugin import get_package, get_inventory
from csmpe.core_plugins.csm_install_operations.utils import get_cmd_for_install_activate_deactivate


class Plugin(CSMPlugin):
    """This plugin deactivates packages on the device."""
    name = "Install Deactivate Plugin"
    platforms = {'ASR9K', 'CRS'}
    phases = {'Deactivate'}
    os = {'XR'}

    def get_tobe_deactivated_pkg_list(self):
        """
        Produces a list of packaged to be deactivated
        """
        packages = self.ctx.software_packages
        pkgs = SoftwarePackage.from_package_list(packages)

        installed_act = SoftwarePackage.from_show_cmd(self.ctx.send("admin show install active summary"))

        if pkgs:
            # packages to be deactivated must be active packages
            packages_to_deactivate = pkgs & installed_act
            if not packages_to_deactivate:
                to_deactivate = " ".join(map(str, pkgs))

                state_of_packages = "\nTo deactivate :{} \nActive: {}".format(
                    to_deactivate, installed_act
                )

                self.ctx.info(state_of_packages)
                self.ctx.error('To be deactivated packages not in inactive packages list.')
                return None
            else:
                if len(packages_to_deactivate) != len(packages):
                    self.ctx.info('Packages selected for deactivation: {}\n'.format(" ".join(map(str, packages))) +
                                  'Packages that are to be deactivated: {}'.format(" ".join(map(str,
                                                                                            packages_to_deactivate))))
                return " ".join(map(str, packages_to_deactivate))

    def run(self):
        """
        Performs install deactivate operation
        """
        cmd = get_cmd_for_install_activate_deactivate(self.ctx, self.get_tobe_deactivated_pkg_list,
                                                      cmd_with_package_names='admin install deactivate {} prompt-level none async',
                                                      cmd_with_operation_id='admin install deactivate id {} prompt-level none async')

        if not cmd:
            self.ctx.info("Nothing to be deactivated.")
            return True

        self.ctx.info("Deactivate package(s) pending")
        self.ctx.post_status("Deactivate Package(s) Pending")

        install_activate_deactivate(self.ctx, cmd)

        self.ctx.info("Deactivate package(s) done")

        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)



# =============================================================================
# delegators
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

from functools import partial


def delegate(attribute_name, method_names, attribute_names=()):
    """Passes the call to the attribute called attribute_name for
    every method listed in method_names.
    """
    # hack for python 2.7 as nonlocal is not available
    d = {
        'delegate': attribute_name,
        'methods': method_names,
        'attributes': attribute_names,
    }

    def setter(attribute, name, instance, value):
        setattr(getattr(instance, attribute), name, value)

    def getter(attribute, name, instance):
        return getattr(getattr(instance, attribute), name)

    def caller(attribute, name):
        return lambda self, *args, **kwargs: getter(attribute, name, self)(*args, **kwargs)

    def decorator(cls):
        delegate_name = d['delegate']
        if delegate_name.startswith("__"):
            delegate_name = "_" + cls.__name__ + delegate_name
        for name in d['methods']:
            setattr(cls, name, caller(delegate_name, name))
        for name in d['attributes']:
            setattr(cls, name, property(partial(getter, delegate_name, name), partial(setter, delegate_name, name)))
        return cls
    return decorator



# =============================================================================
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================
import re


def parse_show_platform(ctx, output):
    """
    :param ctx: PluginContext
    :param output: output from 'show platform'
    :return: dictionary of nodes

    Load for five secs: 1%/0%; one minute: 2%; five minutes: 2%
    No time source, *22:40:38.097 UTC Wed Oct 12 2016
    Chassis type: ASR-920-12CZ-A

    Slot      Type                State                 Insert time (ago)
    --------- ------------------- --------------------- -----------------
     0/0      12xGE-2x10GE-FIXED  ok                    03:07:10
    R0        ASR-920-12CZ-A      ok, active            03:09:23
    F0                            ok, active            03:09:23
    P0        ASR920-PSU0         ok                    never
    P1        ASR920-PSU1         ps, fail              never
    P2        ASR920-FAN          ok                    never

    Slot      CPLD Version        Firmware Version
    --------- ------------------- ---------------------------------------
    R0        1601191C            15.4(3r)S4
    F0        1601191C            15.4(3r)S4


    """

    ctx.info("IOS_XE in parse_show_platform")

    inventory = {}
    lines = output.split('\n')
    lines = [x for x in lines if x]
    slotcnt = 0
    for line in lines:
        line = line.strip()

        # ctx.info("line = {}".format(line))
        # ctx.info("len(line) = {}".format(len(line)))

        if len(line) > 0:

            states = re.split('\s+', line)
            # ctx.info("states[0] = {}".format(states[0]))

            if '--' in states[0]:
                continue

            if states[0] == 'Chassis':
                continue

            if states[0] == 'Slot':
                slotcnt += 1
                if slotcnt == 2:
                    break
                else:
                    continue

            if slotcnt == 1:
                node = line[0:8]
                node = node.strip()
                node_type = line[9:28]
                state = line[29:50]
                state = state.strip()
                config_state = line[51:60]
                config_state = config_state.strip()

                # ctx.info("node={}, type={}, state={}".format(node, node_type, state))

                entry = {
                    'type': node_type,
                    'state': state,
                    'config_state': config_state
                }
                inventory[node] = entry

    return inventory



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================


from csmpe.plugins import CSMPlugin


class Plugin(CSMPlugin):
    """This plugin captures device configuration and stores in the log directory."""
    name = "Config Capture Plugin"
    platforms = {'ASR9K', 'CRS', 'NCS1K', 'NCS4K', 'NCS5K', 'NCS5500', 'NCS6K', 'ASR900', 'N6K', 'IOS-XRv'}
    phases = {'Pre-Upgrade', 'Post-Upgrade'}

    def run(self):
        cmd = "show running-config"
        output = self.ctx.send(cmd, timeout=2200)
        file_name = self.ctx.save_to_file(cmd, output)
        if file_name is None:
            self.ctx.error("Unable to save device configuration to file: {}".format(file_name))
            return False



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================


import re

from csmpe.plugins import CSMPlugin
from install import watch_operation, log_install_errors
from csmpe.core_plugins.csm_get_inventory.exr.plugin import get_package, get_inventory


class Plugin(CSMPlugin):
    """This plugin commits packages on the device."""
    name = "Install Commit Plugin"
    platforms = {'ASR9K', 'NCS1K', 'NCS4K', 'NCS5K', 'NCS5500', 'NCS6K', 'IOS-XRv'}
    phases = {'Commit'}
    os = {'eXR'}

    def run(self):
        """
        It performs commit operation
        RP/0/RP0/CPU0:Deploy#install commit
        May 27 16:34:04 Install operation 32 started by root:
          install commit
        May 27 16:34:05 Install operation will continue in the background

        RP/0/RP0/CPU0:Deploy#May 27 16:34:11 Install operation 32 finished successfully
        """
        cmd = "install commit"
        output = self.ctx.send(cmd)
        result = re.search('Install operation (\d+)', output)
        if result:
            op_id = result.group(1)
            watch_operation(self.ctx, op_id)
        else:
            log_install_errors(self.ctx, output)
            self.ctx.error("Operation ID not found.")
            return

        failed_oper = r'Install operation {} aborted'.format(op_id)
        success_oper = r'Install operation (\d+) finished successfully'

        # Not sure if this is still the message on NCS6K
        completed_with_failure = 'Install operation (\d+) completed with failure'

        cmd = "show install log {} detail".format(op_id)
        output = self.ctx.send(cmd)

        if re.search(failed_oper, output):
            log_install_errors(self.ctx, output)
            self.ctx.error("Install operation failed.")
            return

        if re.search(completed_with_failure, output):
            log_install_errors(self.ctx, output)
            self.ctx.info("Completed with failure but failure was after Point of No Return.")
        elif re.search(success_oper, output):
            self.ctx.info("Operation {} finished successfully.".format(op_id))

        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)



import sys
import importlib


class ServerType:
    TFTP_SERVER = 'TFTP'
    FTP_SERVER = 'FTP'
    SFTP_SERVER = 'SFTP'
    LOCAL_SERVER = 'LOCAL'


def import_module(module, path=None):
    if path is not None:
        sys.path.append(path)
    try:
        return importlib.import_module(module)
    except:
        return None


def concatenate_dirs(dir1, dir2):
    """
    Appends dir2 to dir1. It is possible that either/both dir1 or/and dir2 is/are None
    """
    result_dir = dir1 if dir1 is not None and len(dir1) > 0 else ''
    if dir2 is not None and len(dir2) > 0:
        if len(result_dir) == 0:
            result_dir = dir2
        else:
            result_dir += '/' + dir2

    return result_dir


def is_empty(obj):
    """
    These conditions are considered empty
       s = [], s = None, s = '', s = '    ', s = 'None'
    """
    if isinstance(obj, str):
        obj = obj.replace('None', '').strip()

    if obj:
        return False

    return True


def update_device_info_udi(ctx):
    # _update_device_info() and _update_udi() are removed in condoor-ng
    # ctx._connection._update_device_info()
    # ctx._connection._update_udi()
    ctx._csm.save_data("device_info", ctx._connection.device_info)
    ctx._csm.save_data("udi", ctx._connection.udi)


def get_cmd_for_install_activate_deactivate(ctx, func_get_pkg_list,
                                            cmd_with_package_names, cmd_with_operation_id):
    try:
        has_tar = False
        for package in ctx.software_packages:
            if "tar" in package:
                has_tar = True
                break
        if has_tar:
            operation_id = ctx.get_operation_id(ctx.software_packages)
        else:
            operation_id = None
    except AttributeError:
        ctx.error("No package is selected.")
        return

    if operation_id is None or operation_id == -1:
        if has_tar:
            ctx.error("No install add operation of the exact selected packages has been completed successfully.")
            return
        tobe_activated_deactivated = func_get_pkg_list()
        if not tobe_activated_deactivated:
            return None
        return cmd_with_package_names.format(tobe_activated_deactivated)
    else:
        ctx.info("Using the operation ID: {}".format(operation_id))
        return cmd_with_operation_id.format(operation_id)



# =============================================================================
# migrate.py - plugin for migrating classic XR to eXR/fleXR
#
# Copyright (c)  2013, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import re

from csmpe.plugins import CSMPlugin
from migration_lib import wait_for_final_band, log_and_post_status, run_additional_custom_commands
from csmpe.core_plugins.csm_get_inventory.exr.plugin import get_package, get_inventory
from csmpe.core_plugins.csm_install_operations.utils import update_device_info_udi


XR_PROMPT = re.compile('(\w+/\w+/\w+/\w+:.*?)(\([^()]*\))?#')

SCRIPT_BACKUP_CONFIG_611_DOWN = "harddiskb:/classic.cfg"
SCRIPT_BACKUP_CONFIG_612_UP = "harddiskb:/cXR_xr_plane.cfg"
SCRIPT_BACKUP_ADMIN_CONFIG_611_DOWN = "harddiskb:/admin.cfg"
SCRIPT_BACKUP_ADMIN_CONFIG_612_UP = "harddiskb:/cXR_admin_plane.cfg"

MIGRATION_TIME_OUT = 3600
NODES_COME_UP_TIME_OUT = 3600

PASSWORD_PROMPT = re.compile("[P|p]assword:\s?")
USERNAME_PROMPT = re.compile("([U|u]sername:|login:)\s?")

PERMISSION_DENIED = "Permission denied"
AUTH_FAILED = "Authentication failed|not authorized|Login incorrect"
RESET_BY_PEER = "reset by peer|closed by foreign host"
SET_USERNAME = "[Ee]nter.*username:"
SET_PASSWORD = "Enter secret"
PASSWORD_OK = "[Pp]assword [Oo][Kk]"
PRESS_RETURN = "Press RETURN to get started\."

# Error when the hostname can't be resolved or there is
# network reachability timeout
UNABLE_TO_CONNECT = "nodename nor servname provided, or not known|" \
                    "Unknown host|" \
                    "[Operation|Connection] timed out|" \
                    "[D|d]estination unreachable|" \
                    "[U|u]nable to connect|" \
                    "[C|c]onnection refused"

# Telnet connection initiated
ESCAPE_CHAR = "Escape character is|Open"
# Connection refused i.e. line busy on TS
CONNECTION_REFUSED = re.compile("Connection refused")


class Plugin(CSMPlugin):
    """
    A plugin for migrating a ASR9K IOS-XR(XR) system to
    ASR9K IOS-XR 64 bit(eXR/fleXR).
    This plugin calls the migration script on device and reload
    device to boot eXR image.
    Console access is needed.
    """
    name = "Migrate Plugin"
    platforms = {'ASR9K'}
    phases = {'Migrate'}

    def _run_migration_script(self):
        """
        Run the migration script in /pkg/bin/migrate_to_eXR on device to set
        internal variables for booting eXR image.
        Check that no error occurred.
        :return: True if no error occurred.
        """

        output = self.ctx.send("run /pkg/bin/migrate_to_eXR -m eusb", timeout=600)

        self._check_migration_script_output(output)

        return True

    def _check_migration_script_output(self, output):
        """Check that the migration script finished without errors, and also, the configs are backed up."""
        lines = output.splitlines()
        for line in lines:
            if "No such file" in line:
                self.ctx.error("Found file missing when running migration script. Please check session.log.")
            if "Error:" in line:
                self.ctx.error("Migration script reported error. Please check session.log.")

        if not self._is_config_backed_up(SCRIPT_BACKUP_CONFIG_612_UP, "IOS-XR"):
            if not self._is_config_backed_up(SCRIPT_BACKUP_CONFIG_611_DOWN, "IOS-XR"):
                self.ctx.error("Migration script failed to back up the IOS-XR running config. " +
                               "Please check session.log.")

        if not self._is_config_backed_up(SCRIPT_BACKUP_ADMIN_CONFIG_612_UP, "admin"):
            if not self._is_config_backed_up(SCRIPT_BACKUP_ADMIN_CONFIG_611_DOWN, "admin"):
                self.ctx.error("Migration script failed to back up the admin running config. " +
                               "Please check session.log.")

    def _is_config_backed_up(self, config_filename, config_type):
        output = self.ctx.send('dir {}'.format(config_filename))
        if "No such file" in output:
            return False
        log_and_post_status(self.ctx, "The {} configurations are backed up in {}".format(config_type,
                                                                                         config_filename))
        return True

    def _reload_all(self):
        """Reload all nodes to boot eXR image."""
        if self.ctx.reload(reload_timeout=MIGRATION_TIME_OUT):
            return self._wait_for_reload()
        self.ctx.error("Encountered error when attempting to reload device.")

    def _wait_for_reload(self):
        """Wait for all nodes to come up with max timeout as 18 minutes after the first RSP/RP comes up."""
        log_and_post_status(self.ctx, "Waiting for all nodes to come to FINAL Band.")
        if wait_for_final_band(self.ctx):
            log_and_post_status(self.ctx, "All nodes are in FINAL Band.")
        else:
            log_and_post_status(self.ctx, "Warning: Not all nodes are in FINAL Band after 25 minutes.")

        return True

    def run(self):

        host = None
        try:
            host = self.ctx.get_host
        except AttributeError:
            self.ctx.error("No host selected.")

        if host is None:
            self.ctx.error("No host selected.")

        log_and_post_status(self.ctx,
                            "Run migration script to extract the image and boot files and set boot mode in device")

        self._run_migration_script()

        log_and_post_status(self.ctx, "Reload device to boot ASR9K-X64 image.")
        self._reload_all()

        run_additional_custom_commands(self.ctx, {"show platform"})

        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)

        update_device_info_udi(self.ctx)

        return True



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================
import re
import time

from csmpe.core_plugins.csm_node_status_check.ios_xe.plugin_lib import parse_show_platform
from utils import install_add_remove

plugin_ctx = None


def send_newline(fsm_ctx):
    fsm_ctx.ctrl.sendline('\r\n')
    fsm_ctx.ctrl.sendline('\r\n')
    fsm_ctx.ctrl.sendline('\r\n')
    return True


def issu_error_state(fsm_ctx):
    plugin_ctx.warning("Error in ISSU. Please see session.log for details")
    return False


def issu_connection_closed(fsm_ctx):
    plugin_ctx.warning("Unexpected connection closed by foreign host during ISSU")
    # sleep until both standby and active RSP's are upgraded
    time.sleep(3600)
    return True


def validate_node_state(inventory):
    valid_state = [
        'ok',
        'ok, active',
        'ok, standby',
        'ps, fail',
        'out of service',
        'N/A'
    ]
    for key, value in inventory.items():
        if value['state'] not in valid_state:
            break
    else:
        return True
    return False


def wait_for_reload(ctx):
    """
     Wait for system to come up with max timeout as 25 Minutes

    """
    begin = time.time()
    ctx.disconnect()
    ctx.post_status("Waiting for device boot to reconnect")
    ctx.info("Waiting for device boot to reconnect")
    time.sleep(1500)   # 25 * 60 = 1500
    ctx.reconnect(force_discovery=True)
    ctx.info("Boot process finished")
    ctx.info("Device connected successfully")

    timeout = 3600
    poll_time = 30
    time_waited = 0

    ctx.info("Waiting for all nodes to come up")
    ctx.post_status("Waiting for all nodes to come up")
    time.sleep(30)

    output = None

    ncnt = 0
    while 1:

        ncnt += 1
        if ncnt > 20:
            break

        # Wait till all nodes are in XR run state
        time_waited += poll_time
        if time_waited >= timeout:
            break

        time.sleep(poll_time)

        # show platform can take more than 1 minute after router reload. Issue No. 47
        output = ctx.send('show platform', timeout=600)

        ctx.info("show platform = {}".format(output))

        inventory = parse_show_platform(ctx, output)
        if validate_node_state(inventory):
            ctx.info("All nodes in desired state")
            elapsed = time.time() - begin
            ctx.info("Overall outage time: {} minute(s) {:.0f} second(s)".format(elapsed // 60, elapsed % 60))
            return True

    # Some nodes did not come to run state
    ctx.error("Not all nodes have came up: {}".format(output))
    # this will never be executed
    return False


def install_activate_write_memory(ctx, cmd):
    """

    PAN-5201-ASR903#write memory
    Building configuration...
    [OK]
    PAN-5201-ASR903#

    PAN-5201-ASR903#write memory
    Warning: Attempting to overwrite an NVRAM configuration previously written
    by a different version of the system image.
    Overwrite the previous NVRAM configuration?[confirm]

    """
    global plugin_ctx
    plugin_ctx = ctx

    # Seeing this message without the reboot prompt indicates a non-reload situation
    Build_config = re.compile("\[OK\]")
    Overwrite_warning = re.compile("Overwrite the previous NVRAM configuration\?\[confirm\]")

    events = [Overwrite_warning, Build_config]
    transitions = [
        (Overwrite_warning, [0], 1, send_newline, 1200),
        (Build_config, [0, 1], -1, None, 1200),
    ]

    if not ctx.run_fsm("write memory", cmd, events, transitions, timeout=1200):
        ctx.error("Failed: {}".format(cmd))


def expand_subpkgs_exec(ctx, folder, pkg):
    """
    Expand the consolidated file into the image folder

    :param: ctx
    :param: folder
    :param: pkg ie bootflash:asr900*.bin
    :return: True or False
    """
    pkg_conf = folder + '/packages.conf'
    pkg_conf2 = pkg_conf + '-'

    output = ctx.send('dir ' + pkg_conf)
    if 'No such file or directory' not in output:
        ctx.send('del /force ' + pkg_conf2)
        cmd = 'copy ' + pkg_conf + ' ' + pkg_conf2
        install_add_remove(ctx, cmd)
        ctx.send('del /force ' + pkg_conf)

    ctx.info("Expanding subpackages into {}".format(folder))

    cmd = 'request platform software package expand file ' + pkg + ' to ' + folder
    output = ctx.send(cmd, timeout=600)
    m = re.search('SUCCESS: Finished expanding all-in-one software package', output)
    if not m:
        ctx.warning("Error: {}".format(cmd))
        return False

    return True


def expand_subpkgs(ctx, rsp_count, folder, pkg):
    """
    Expand the consolidated file into the image folder

    :param: ctx
    :param: rsp_count
    :param: pkg
    :return: True or False
    """

    package = 'bootflash:' + pkg
    result = expand_subpkgs_exec(ctx, folder, package)
    if not result:
        ctx.error('Expanding {} into {} has encountered '
                  'an error'.format(package, folder))
        return False

    if rsp_count == 2:
        cmd = 'copy bootflash:' + pkg + ' ' + 'stby-bootflash:' + pkg
        install_add_remove(ctx, cmd)

        package = 'stby-' + package
        folder = 'stby-' + folder
        result = expand_subpkgs_exec(ctx, folder, package)
        if not result:
            ctx.error('Expanding {} into {} has encountered '
                      'an error'.format(package, folder))
            return False

    return True


def install_activate_reload(ctx):
    """
    Reload the router

    :param ctx
    :return: nothing
    """
    message = "Waiting the {} operation to continue".format('reload')
    ctx.info(message)
    ctx.post_status(message)

    if not ctx.reload(reload_timeout=1200, no_reload_cmd=True):
        ctx.error("Encountered error when attempting to reload device.")

    success = wait_for_reload(ctx)

    if not success:
        ctx.error("Reload or boot failure")
        return

    ctx.info("Operation reload finished successfully")
    return


def install_activate_issu(ctx, cmd):
    """
    Start the issu

    :param ctx
    :param cmd
    :param hostname
    :return: nothing
    """

    global plugin_ctx
    plugin_ctx = ctx

    # Seeing a message without STAGE 4 is an error
    Phase_one = re.compile("Starting disk space verification")
    Stage_one = re.compile("STAGE 1: Installing software on standby RP")
    Stage_two = re.compile("STAGE 2: Restarting standby RP")
    Stage_three = re.compile("STAGE 3: Installing sipspa package on local RP")
    Stage_four = re.compile("STAGE 4: Installing software on active RP")
    Load_on_reboot = re.compile("SUCCESS: Software provisioned.  New software will load on reboot")
    Missing_conf = re.compile("SYSTEM IS NOT BOOTED VIA PACKAGE FILE")
    Failed = re.compile("FAILED:")
    Connection_closed = re.compile("Connection closed by foreign host")
    #            0          1          2          3             4            5
    events = [Phase_one, Stage_one, Stage_two, Stage_three, Stage_four, Load_on_reboot,
              Missing_conf, Failed, Connection_closed]
    #            6            7            8
    transitions = [
        (Phase_one, [0], 1, None, 1800),
        (Stage_one, [0, 1], 2, None, 1800),
        (Stage_two, [2], 3, None, 1800),
        (Stage_three, [3], 4, None, 1800),
        (Stage_four, [4], 5, None, 1800),
        (Load_on_reboot, [5], -1, None, 1800),
        (Missing_conf, [0, 1, 2, 3, 4, 5], -1, issu_error_state, 60),
        (Failed, [0, 1, 2, 3, 4, 5], -1, issu_error_state, 60),
        (Connection_closed, [1, 2, 3, 4], -1, issu_connection_closed, 4200)
    ]

    if not ctx.run_fsm("ISSU", cmd, events, transitions, timeout=4200):
        ctx.error("Failed: {}".format(cmd))

    time.sleep(300)

    success = wait_for_reload(ctx)

    if not success:
        ctx.error("Reload or boot failure")
        return

    ctx.info("Operation reload finished successfully")
    return



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import time
from csmpe.plugins import CSMPlugin


class Plugin(CSMPlugin):
    """This plugin retrieves software information from the device."""
    name = "Get Inventory Plugin"
    platforms = {'ASR9K', 'NCS1K', 'NCS4K', 'NCS5K', 'NCS5500', 'NCS6K', 'IOS-XRv'}
    phases = {'Get-Inventory'}
    os = {'eXR'}

    def run(self):
        get_package(self.ctx)
        get_inventory(self.ctx)


def get_inventory(ctx):
    # Save the output of "show inventory" in admin mode
    output = get_output_in_admin_mode(ctx, "show inventory")
    ctx.save_data("cli_show_inventory", output)


def get_package(ctx):
    """
    Convenient method, it may be called by outside of the plugin
    """

    # Get the admin packages
    ctx.save_data("cli_admin_show_install_inactive",
                  get_output_in_admin_mode(ctx, "show install inactive"))
    ctx.save_data("cli_admin_show_install_active",
                  get_output_in_admin_mode(ctx, "show install active"))
    ctx.save_data("cli_admin_show_install_committed",
                  get_output_in_admin_mode(ctx, "show install committed"))

    # Get the non-admin packages
    ctx.save_data("cli_show_install_inactive",
                  get_output_in_admin_mode(ctx, "show install inactive", admin=False))
    ctx.save_data("cli_show_install_active",
                  get_output_in_admin_mode(ctx, "show install active", admin=False))
    ctx.save_data("cli_show_install_committed",
                  get_output_in_admin_mode(ctx, "show install committed", admin=False))


def get_output_in_admin_mode(ctx, cmd, admin=True):
    """
    :param ctx:
    :param cmd:
    :param admin: True - Calvados, False - xr
    :return: cmd ouput

    Polling as a workaround when the router cmd is not ready:

    RP/0/RP0/CPU0:xrg-ncs-2#show install active
    Node 0/0/CPU0
    Node unresponsive (possible ongoing install operation).
    Please try command later

    Node 0/2/CPU0
    Node unresponsive (possible ongoing install operation).
    Please try command later
    """

    if admin:
        command = 'admin ' + cmd
        ctx.send("admin")
    else:
        command = cmd

    x = 0
    output = ctx.send(cmd)
    while x < 60:
        if 'Please try command later' in output:
            x += 1
            time.sleep(10)
            output = ctx.send(cmd)
        else:
            break

    if admin:
        ctx.send("exit")

    if 'Please try command later' in output:
        ctx.warning('The command {} is not ready after 10 minutes. Please manually '
                    'retrieve latest software from the Host Dashboard'.format(command))

    return output



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================


import re
from csmpe.plugins import CSMPlugin
from utils import available_space
from utils import number_of_rsp
from utils import install_folder
from utils import check_issu_readiness
from utils import remove_exist_subpkgs
from utils import install_package_family
from utils import create_folder
from utils import xe_show_platform
from utils import check_pkg_conf


class Plugin(CSMPlugin):
    """This plugin performs pre-activate tasks."""
    name = "Install Pre-Activate Plugin"
    platforms = {'ASR900'}
    phases = {'Pre-Activate'}
    os = {'XE'}

    def run(self):

        self.ctx.info("Hardware platform: {}".format(self.ctx._connection.platform))
        self.ctx.info("OS Version: {}".format(self.ctx._connection.os_version))

        try:
            packages = self.ctx.software_packages
        except AttributeError:
            self.ctx.warning("No package list provided. Skipping calculation of required free bootflash memory.")
            return

        pkg = ''.join(packages)
        con_platforms = ['ASR-902', 'ASR-920']
        sub_platforms = ['ASR-903', 'ASR-907']
        rsp_count = 1
        folder = 'bootflash:'
        stby_folder = 'stby-bootflash:'

        # check the device type vs the package family
        supported_imgs = {}
        supported_imgs['asr902'] = ['asr900', 'asr903']
        supported_imgs['asr903'] = ['asr900', 'asr903']
        supported_imgs['asr907'] = ['asr900', 'asr903']
        supported_imgs['asr920'] = ['asr920']

        m = re.search('ASR-(\d+)', self.ctx._connection.platform)
        if m:
            device_family = m.group(1)
            device_family = 'asr' + device_family
        else:
            self.ctx.error("Unspported device: {}".format(self.ctx._connection.platform))
            return

        pkg_family = install_package_family(pkg)

        if not pkg_family:
            self.ctx.info("Private device image: {}".format(pkg))

        if pkg_family not in supported_imgs[device_family]:
            self.ctx.info("Private device image: {} on {}".format(pkg, self.ctx._connection.platform))

        # check the RSP type between image and device:
        curr_rsp = None
        pkg_rsp = None

        output = self.ctx.send("show version | include RSP")
        if output:
            m = re.search('(RSP\d)', output)
            if m:
                curr_rsp = m.group(0).lower()

            m = re.search('(rsp\d)', pkg)
            if m:
                pkg_rsp = m.group(0)

            if curr_rsp and pkg_rsp and curr_rsp != pkg_rsp:
                self.ctx.info("Incompatible Route processor in {} for this device {}".format(pkg, curr_rsp))

        # Determine one of the following modes: consolidated, subpackage, or issu

        if self.ctx._connection.platform in con_platforms:
            mode = 'consolidated'
        elif self.ctx._connection.platform in sub_platforms:
            mode = 'subpackage'
            # Determine the number of RSP's in the chassis
            rsp_count = number_of_rsp(self.ctx)
            if rsp_count == 0:
                self.ctx.error("No RSP is discovered")
                return

            # Determine the install folder
            folder = install_folder(self.ctx)
            stby_folder = 'stby-' + folder

            # Create the folder if it does not exist
            if not create_folder(self.ctx, folder):
                self.ctx.error("Install folder {} creation failed", format(folder))
                return

            if rsp_count == 2 and not create_folder(self.ctx, stby_folder):
                self.ctx.error("Install folder {} creation "
                               "failed", format(stby_folder))
                return
        else:
            self.ctx.error("Unsupported platform: {}".format(self.ctx._connection.platform))
            return

        total_size = 10000000

        valid_pkg_conf = False
        if mode == 'subpackage':
            # Check if the packages.conf is valid
            valid_pkg_conf = check_pkg_conf(self.ctx, folder)

            # Remove residual image files from previous installations
            if valid_pkg_conf:
                remove_exist_subpkgs(self.ctx, folder, pkg)
            else:
                self.ctx.warning("Empty or invalid {}/packages.conf".format(folder))
                self.ctx.warning("Residual packages from previous installations are not "
                                 "automatically removed from bootflash: / stby-bootflash:.")
                self.ctx.info("Sub-package mode will be performed to "
                              "activate package = {}".format(pkg))

            cmd = "dir bootflash: | include " + pkg
            output = self.ctx.send(cmd)
            if output:
                m = re.search('-rw-\s+(\d+)\s+', output)
                if m:
                    total_size += int(m.group(1))

        flash_free = available_space(self.ctx, 'bootflash:')
        self.ctx.info("Total required / bootflash "
                      "available: {} / {} bytes".format(total_size, flash_free))
        if flash_free < total_size:
            self.ctx.error("Not enough space on bootflash: to install packages. "
                           "The install process can't proceed.\n"
                           "Please erase unused images, crashinfo, "
                           "core files, and tracelogs")
        else:
            self.ctx.info("There is enough space on bootflash: to install packages.")

        if rsp_count == 2:
            if valid_pkg_conf:
                remove_exist_subpkgs(self.ctx, stby_folder, pkg)
            stby_free = available_space(self.ctx, 'stby-bootflash:')
            self.ctx.info("Total required / stby-bootflash "
                          "available: {} / {} bytes".format(total_size, stby_free))
            if stby_free < total_size:
                self.ctx.error("Not enough space on stby-bootflash: to "
                               "install packages. The install process can't proceed.\n"
                               "Please erase unused images, crashinfo, core files, "
                               "and tracelogs")
            else:
                self.ctx.info("There is enough space on stby-bootflash: to install packages.")

        # Determine if ISSU is feasible
        if mode == 'subpackage' and rsp_count == 2 and valid_pkg_conf:
            if check_issu_readiness(self.ctx, pkg, total_size):
                mode = 'issu'
                self.ctx.info("ISSU will be performed to activate package = {}".format(pkg))

        # Log the status of RP and SIP
        platform_info = xe_show_platform(self.ctx)
        if not platform_info:
            self.ctx.error("The CLI 'show platform' is not able to determine the status of RP and SIP ")
            return
        self.ctx.info("show platform = {}".format(platform_info))

        self.ctx.info("Activate number of RSP = {}".format(rsp_count))
        self.ctx.info("Activate package = {}".format(pkg))
        self.ctx.info("Install folder = {}".format(folder))
        self.ctx.info("Activate package mode = {}".format(mode))

        self.ctx.save_data('xe_rsp_count', rsp_count)
        self.ctx.save_data('xe_activate_pkg', pkg)
        self.ctx.save_data('xe_boot_mode', mode)
        self.ctx.save_data('xe_install_folder', folder)
        self.ctx.save_data('xe_show_platform', platform_info)

        return True



# =============================================================================
# Context
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import logging
import os
import re
from time import time

import condoor
from decorators import delegate


class PluginError(Exception):
    pass


class InstallContext(object):
    _storage = {}

    def __init__(self):
        self.hostname = "Hostname"
        self._custom_commands = []

    def post_status(self, message):
        print("[CSM Status] {}".format(message))

    def save_data(self, key, value):
        #  print("Saving [{}]={}".format(key, str(value[0])))
        self._storage[key] = value

    def load_data(self, key):
        #  print("Loading [{}]".format(key))
        return self._storage.get(key, (None, None))

    @property
    def custom_commands(self):
        return self._custom_commands

    @custom_commands.setter
    def custom_commands(self, value):
        if isinstance(value, str):
            value = list(value)
        self._custom_commands = value


class Host(object):
    pass


@delegate("_csm", ("post_status",), ("custom_commands", "success", "get_operation_id", "set_operation_id",
                                     "server_repository_url", "software_packages", "hostname", "log_directory",
                                     "migration_directory", "get_server", "get_host"))
@delegate("_connection", ("connect", "disconnect", "reconnect", "discovery", "send", "run_fsm", "reload"),
          ("family", "prompt", "os_type", "os_version", "is_console"))
class PluginContext(object):
    """ This is a class passed to the constructor during plugin instantiation.
    Thi class provides the API for the plugins to allow the communication with the CMS Server and device.
    """
    def __init__(self, csm=None):
        self._csm = csm
        self._log_handler = None
        self.current_plugin = ""
        if csm is not None:
            self._connection = condoor.Connection(
                self._csm.hostname,
                self._csm.host_urls,
                log_dir=self._csm.log_directory
            )
            self._set_logging(hostname=self._csm.hostname, log_dir=self._csm.log_directory, log_level=logging.DEBUG)
            self._connection.msg_callback = self._post_and_log
            self._connection.error_msg_callback = self._error_callback

            self._device_detect()
        else:
            self._connection = None
            self._set_logging()

    def _post_and_log(self, message):
        self.info(message)
        self.post_status(message)
        return

    def _error_callback(self, message):
        self.save_job_info('ERROR:' + self._format_log(message))
        """Log ERROR message"""
        self._logger.error(self._format_log(message))

    def _set_logging(self, hostname="host", log_dir=None, log_level=logging.NOTSET):
        self._logger = logging.getLogger("{}.plugin_manager".format(hostname))
        formatter = logging.Formatter('%(asctime)-15s %(levelname)8s: %(message)s')
        if log_dir:
            if not os.path.exists(log_dir):
                try:
                    os.makedirs(log_dir)
                except IOError:
                    log_dir = "./"
            log_filename = os.path.join(log_dir, 'plugins.log')
            handler = logging.FileHandler(log_filename)

        else:
            handler = logging.StreamHandler()

        handler.setFormatter(formatter)
        self._log_handler = handler
        self._logger.addHandler(handler)
        self._logger.setLevel(log_level)

    def _reset_logging(self):
        self._logger.removeHandler(self._log_handler)
        self._log_handler.close()

    def __del__(self):
        self.finalize()

    def finalize(self):
        """Clean up the the context."""
        if self._connection:
            self._connection.disconnect()
            self._connection.finalize()

        self._reset_logging()

    @property
    def TIMEOUT(self):
        return condoor.TIMEOUT

    @property
    def CommandTimeoutError(self):
        return condoor.CommandTimeoutError

    @property
    def phase(self):
        """
        :return: string representing the installation phase.
        """
        try:
            return self._csm.requested_action
        except AttributeError:
            pass
            # raise AssertionError("Requested action not provided")

    def _device_detect(self):
        """Connect to device using condoor"""
        self.info("Phase: Connecting")
        self.post_status("Connecting to device")

        # Use force discovery until we sort out all cache issues.
        self.connect(force_discovery=True)
        """
        if self.phase in ["Get-Inventory", "Post-Upgrade"]:
            self.connect(force_discovery=True)
        else:
            self.connect()
        """

        self.info("Hostname: {}".format(self._connection.hostname))
        self.info("Hardware family: {}".format(self._connection.family))
        self.info("Hardware platform: {}".format(self._connection.platform))
        self.info("OS type: {}".format(self._connection.os_type))
        self.info("Version: {}".format(self._connection.os_version))
        self.info("Connection type: {}".format("console" if self._connection.is_console else "vty"))

        self._csm.save_data("device_info", self._connection.device_info)
        self._csm.save_data("udi", self._connection.udi)

    def _format_log(self, message):
        return "[{}] {}".format(self.current_plugin, message) if self.current_plugin else "{}".format(message)

    def info(self, message):
        """Log INFO message"""
        self._logger.info(self._format_log(message))

    def error(self, message):
        self.save_job_info('ERROR:' + self._format_log(message))

        """Log ERROR message"""
        self._logger.error(self._format_log(message))
        self.finalize()
        raise PluginError

    def warning(self, message):
        self.save_job_info('WARNING: ' + self._format_log(message))

        """Log WARNING message"""
        self._logger.warning(self._format_log(message))

    def save_job_info(self, message):
        try:
            self._csm.save_job_info(message)
        except AttributeError:
            pass

    # Storage API
    def save_data(self, key, data):
        """
        Stores (data, timestamp) tuple for key adding timestamp
        This tuple is saved to host context data
        """
        self._csm.save_data(key, [data, time()])
        self.info("Key '{}' saved in CSM storage".format(key))

    def load_data(self, key):
        """
        Loads (data, timestamp) tuple for the key from host context data
        """
        result = self._csm.load_data(key)
        if result:
            self.info("Key '{}' loaded from CSM storage".format(key))
            if isinstance(result, list):
                return tuple(result)
            else:
                return result, None
        return None, None

    # Storage API
    def save_job_data(self, key, data):
        """
        Stores (data, timestamp) tuple for key adding timestamp
        This tuple is saved to install job data
        """
        self._csm.save_job_data(key, [data, time()])
        self.info("Key '{}' saved in CSM storage".format(key))

    def load_job_data(self, key):
        """
        Loads (data, timestamp) tuple for the key
        """
        result = self._csm.load_job_data(key)
        if result:
            self.info("Key '{}' loaded from CSM storage".format(key))
            if isinstance(result, list):
                return tuple(result)
            else:
                return result, None
        return None, None

    def normalize_filename(self, name):
        filename = re.sub(r"\W+", '-', name)
        filename += ".txt"
        return filename

    def save_to_file(self, name, data):
        """
        Save data to filename in the log_directory provided by CSM
        """

        store_dir = self._csm.log_directory
        file_name = self.normalize_filename(name)
        full_path = os.path.join(store_dir, file_name)
        with open(full_path, "w") as f:
            f.write(data)
            self.info("File '{}' saved in CSM log directory".format(file_name))
            return file_name
        return None

    def load_from_file(self, file_name):
        """
        Load data from file where full path is provided as file_name
        """
        full_path = file_name
        with open(full_path, "r") as f:
            data = f.read()
            self.info("File '{}' loaded from CSM directory".format(os.path.basename(file_name)))
            return data
        return None
